/* fontanim.c:
 * fontanim v0.9 Copyright (c) November, 1997 by Greg Alexander, to be
 * distributed under the terms of the GNU General Public License, included
 * in the file COPYING distributed with this program.
 */
 
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <linux/kd.h>
#include <sys/ioctl.h>
#include <sys/termios.h>

#define SIZE 8192	/* Size of one full 256-char font record */

struct animrec {
	int start_font,num_fonts;
	int x,y;
	int num_frames;
	unsigned char **frames;
	struct animrec *next;
} *animroot=NULL;

unsigned int start_font=128;
unsigned char frows;

int load_anim(char *fn);
void get_font(unsigned char *s);
void set_font(unsigned char *s);
int draw_stuff(struct animrec *t,int rows,int cols);
void animate(struct animrec *root,int delay);

int usage(char *argv0)
{
	fprintf(stderr,"Usage: %s file1.anim ...\n",argv0);
	return 1;
}

int main(int argc, char *argv[])
{
	int delay=100000;
	int i;
	struct winsize w;
	struct consolefontdesc cfd;
	
	if (ioctl(1,TIOCGWINSZ,&w))
		return perror("ioctl(TIOCGWINSZ)"),-1;

	cfd.chardata=NULL;
	if (ioctl(1,GIO_FONTX,&cfd))
		return perror("ioctl(GIO_FONTX)"),-1;
	frows=cfd.charheight;

	if (fork())
		return 0;
	
	if (argc<2)
		return usage(argv[0]);

	for (i=1;i<argc;i++)
		start_font+=load_anim(argv[i]);
	if (start_font>255)
		return fprintf(stderr,"Too many fonts required!\n"),1;

	if (!draw_stuff(animroot,w.ws_row,w.ws_col))
		return 1;
	do {
		animate(animroot,delay);
	} while (animroot->num_frames>1);
		
	return 0;
}

int load_anim(char *fn)
{
	char s[1000];
	struct animrec *n=malloc(sizeof(struct animrec));
	struct animrec **t=&animroot;
	FILE *f;
	int i,nx,ny,x,y;
	
	if (!n)
		return fprintf(stderr,"No mem!\n"),0;
	f=fopen(fn,"r");
	if (!f)
		return perror(fn),free(n),0;
	fgets(s,1000,f);
	sscanf(s,"%d %d %d",&(n->num_frames),&nx,&ny);
	n->x=(nx-1)/8+1;
	n->y=(ny-1)/frows+1;
	n->next=NULL;
	n->start_font=start_font;
	n->num_fonts=n->x*n->y;
	n->frames=malloc(sizeof(unsigned char *)*n->num_frames);
	
	for (i=0;i<n->num_frames;i++)
	{
		if (!fgets(s,1000,f))
			return fprintf(stderr,"Premature EOF!\n"),free(n->frames),free(n),0;
		strtok(s,"\r\n");
		if (*s=='-') {
			n->frames[i]=NULL;
			continue;
		}		
		n->frames[i]=malloc(sizeof(unsigned char)*frows*n->num_fonts);
		if (!n->frames[i])
			return fprintf(stderr,"No mem!\n"),free(n->frames),free(n),0;
		memset(n->frames[i],0,sizeof(unsigned char)*frows*n->num_fonts);
		for (y=0;y<ny;y++) {
			if (!fgets(s,1000,f))
				return fprintf(stderr,"Premature EOF!\n"),free(n->frames),free(n),0;
			strtok(s,"\r\n");
			if (strlen(s)>nx)
				fprintf(stderr,"Warning: too many (%d) characters in line: \"%s\"!\n",strlen(s),s);
			if (strlen(s)<nx)
				return fprintf(stderr,"Error: too few (%d) characters in line: \"%s\"!\n",strlen(s),s),free(n->frames),free(n),0;
			for (x=0;x<nx;x++)
				n->frames[i][((x/8)+y/frows*n->x)*frows + y%frows] |= (s[x]-'0')<<(7-(x%8));
		}
	}
	
	for (;*t;t=&((*t)->next))
		;
	*t=n;
	return n->num_fonts;
}

void get_font(unsigned char *s)
{
	ioctl(2,GIO_FONT,s);
}

void set_font(unsigned char *s)
{
	ioctl(2,PIO_FONT,s);
}

int draw_stuff(struct animrec *t,int rows,int cols)
{
	int x,y,i,xx=0;
	FILE *f=fopen("/dev/vcs0","w");
	
	if (!f)
		return 0;
	
	while (t)
	{
		i=t->start_font;
		for (y=0;y<t->y;y++) {
			fseek(f,(rows-1-t->y+y)*cols+xx,SEEK_SET);
			for (x=0;x<t->x;x++)
				fprintf(f,"%c",i++);
		}
		xx+=t->x;
		t=t->next;
	}
	fclose(f);
	return 1;
}

void animate(struct animrec *root,int delay)
{
	struct animrec *t;
	unsigned char font[SIZE];
	int i,j,done=0;
	
	if (!root)
		return;

	get_font(font);
	
	for (i=0;!done;i++) {
		t=root;
		done=1;
		do {
			if (i>=t->num_frames)
				continue;
			done=0;
			for (j=0;j<t->num_fonts;j++)
				memcpy(font+(j+t->start_font)*32,t->frames[i]+frows*j,frows);
		} while ((t=t->next));
		set_font(font);
		usleep(delay);
	}
}
